/*
    mpg321 - a fully free clone of mpg123.
    Copyright (C) 2001 Joe Drew
    
    Originally based heavily upon:
    plaympeg - Sample MPEG player using the SMPEG library
    Copyright (C) 1999 Loki Entertainment Software
    
    Also uses some code from
    mad - MPEG audio decoder
    Copyright (C) 2000-2001 Robert Leslie
    
    Original playlist code contributed by Tobias Bengtsson <tobbe@tobbe.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define FAKEVERSION "0.59q"
#define VERSIONDATE "2001/Aug/08"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <signal.h>
#include <limits.h>
#include <sys/time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "getopt.h" /* GNU getopt is needed, so I included it */
#include "mpg321.h"

/* temporary globals */
int shuffle_play;
int stop_playing_file = 0;
int quit_now = 0;
char *playlist_file;
ao_device_t *playdevice=NULL;
mad_timer_t current_time;
mpg321_options options = { 0, NULL, 0 };

void mpg123_boilerplate()
{
    fprintf(stderr,"High Performance MPEG 1.0/2.0/2.5 Audio Player for Layer 1, 2, and 3.\n"
        "Version " FAKEVERSION " (" VERSIONDATE "). Written and copyrights by Joe Drew.\n"
        "Uses code from various people. See 'README' for more!\n"
        "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n");
}

void usage(char *argv0)
{
    mpg123_boilerplate();
    fprintf(stderr,
        "\nUsage: %s [options] file(s) | URL(s) | -\n\n"
        "Options supported:\n"
        "    --verbose or -v              Increase verbosity\n"
        "    --quiet or -q                Quiet mode (no title or boilerplate)\n"
        "    --gain N or -g N             Set gain (audio volume) to N (0-100)\n"
        "    --skip N or -k N             Skip N frames into the file\n"
        "    --verbose or -v              Be more verbose in playing files\n"
        "    -o dt                        Set output devicetype to dt (esd or alsa)\n"
        "    --audiodevice N or -a N      Use N for audio-out\n"
        "    --stdout or -s               Use stdout for audio-out\n"
        "    --wav N or -w N              Use wave file N for output\n"
        "    --test or -t                 Test only; do no audio output\n"
        "    --list N or -@ N             Use playlist N as list of MPEG files\n"
        "    --random or -Z               Play files randomly until interrupted\n"
        "    --shuffle or -z              Shuffle list of files before playing\n"
        "    --help or --longhelp         Print this help screen\n"
        "    --version or -V              Print version information\n" , argv0);
}

/* retsigtype is defined by configure;
   pass handle_signals -1 to initialize it */
RETSIGTYPE handle_signals(int sig) 
{
    static struct timeval last_signal = { 0, 0 };
    struct timeval curr_tv;
    
    gettimeofday(&curr_tv, NULL);
    
    if (sig != -1)
    {
        if (((curr_tv.tv_sec - last_signal.tv_sec) * 1000000 +  
            (curr_tv.tv_usec - last_signal.tv_usec)) < 750000) /* Allow a short delay to still
                                                                  kill mpg321 */
        {
            quit_now = 1;
        }
            
        last_signal.tv_sec = curr_tv.tv_sec;
        last_signal.tv_usec = curr_tv.tv_usec;
        
        stop_playing_file = 1;
   }
   else /* initialize */
   {
        last_signal.tv_sec = curr_tv.tv_sec;
        last_signal.tv_usec = curr_tv.tv_usec;
   }
}

void parse_options(int argc, char *argv[], playlist *pl)
{
    struct option long_options[] =
    {
        /* NO-OPS. Implement these if you need them, please. */
    
        /* these take no parameter and have no short equiv */
        { "headphones", 0, 0, 'O' }, /* or -o h */
        { "speaker", 0, 0, 'P' }, /* or -o s */
        { "lineout", 0, 0, 'L' }, /* or -o l */
        { "stereo", 0, 0, 'T' },
        { "reopen", 0, 0, 'R' },
        { "equalizer", 0, 0, 'E' },
        { "aggressive", 0, 0, 'A' },
        { "8bit", 0, 0, '8' },
    
        /* these take no parameter and have short equiv */
        { "2to1", 0, 0, '2' },
        { "4to1", 0, 0, '4' },
        { "check", 0, 0, 'c' },
        { "resync", 0, 0, 'y' },
        { "left", 0, 0, '0' },
        { "single0", 0, 0, '0' },
        { "right", 0, 0, '1' },
        { "single1", 0, 0, '1' },
        { "mono", 0, 0, 'm' },
        { "mix", 0, 0, 'm' },
        { "control", 0, 0, 'C' },
        { "auth", 0, 0, 'u' },
    
        /* These take a parameter and have no short equiv */
        { "au", 1, 0, 'U' },
        { "cdr", 1, 0, 'D' },
    
        /* these take a parameter and have short equiv */
        { "frames", 1, 0, 'n' },
        { "doublespeed", 1, 0, 'd' },
        { "halfspeed", 1, 0, 'h' },
        { "scale", 1, 0, 'f' },
        { "buffer", 1, 0, 'b' },
        { "proxy", 1, 0, 'p' },
        { "rate", 1, 0, 'r' },
            
        /* The following are all implemented. */
            
        /* Takes no parameters */
        { "verbose", 0, 0, 'v' },
        { "quiet", 0, 0, 'q' }, 
        { "test", 0, 0, 't' },
        { "stdout", 0, 0, 's' },
	{ "STDOUT", 0, 0, 's' },
        { "version", 0, 0, 'V' },
        { "help", 0, 0, 'H' },
        { "longhelp", 0, 0, 'H' },
        { "shuffle", 0, 0, 'z' },
        { "random", 0, 0, 'Z' },
            
        /* takes parameters */
        { "output", 1, 0, 'o' },
        { "list", 1, 0, '@' },
        { "skip", 1, 0, 'k' },
        { "wav", 1, 0, 'w' },
        { "audiodevice", 1, 0, 'a' },
        { "gain", 1, 0, 'g' },
        { 0, 0, 0, 0 }
    };
    int option_index = 0, c;

    while ((c = getopt_long(argc, argv, "24cy01m8CzZun:d:h:f:b:p:@:r:k:w:a:g:vqtlsVo:", 
                        long_options, &option_index)) != -1)
    {            
        switch(c)
        {
            case 'O': case 'P': case 'L': case 'T': case 'R': case 'E': case 'A': case '8':
            case '2': case '4': case 'c': case 'y': case '0': case '1': case 'm': case 'C':
            case 'u':
            case 'U': case 'D': case 'n': case 'd': case 'h': case 'f': case 'b': case 'p':
            case 'r':
                break;

            case 'z':
                shuffle_play = 1;
                break;
                
            case 'Z':
                set_random_play(pl);
                break;
                
            case '@': 
                playlist_file = strdup(optarg);
                break;

            case 'v':
                options.opt |= MPG321_VERBOSE_PLAY;
#ifndef SETVBUF_REVERSED
                setvbuf(stdout, NULL, _IONBF, 0);
#else
		setvbuf(stdout, _IONBF, NULL, 0);
#endif		
                break;
            
            case 'q':
                options.opt |= MPG321_QUIET_PLAY;
                break;
            
            case 'k':
                options.seek = atol(optarg);
                if (options.seek < 0)
                {
                    fprintf(stderr, "Number of frames to skip must be positive!\n");
                    exit(1);
                }
                break;
            
            case 't':
                options.opt |= MPG321_USE_NULL;
                break;
            
            case 'w':
                options.opt |= MPG321_USE_WAV;
                options.device = strdup(optarg);
                break;
            
            case 'a':
                /* use default output device, or whatever device is specified with -o */
                options.device = strdup(optarg);
                break;
            
            case 's':
                options.opt |= MPG321_USE_STDOUT;
                break;
                
            case 'o':
                if (strcmp(optarg, "alsa") == 0)
                {
                    options.opt |= MPG321_USE_ALSA;
                }
                
                else if (strcmp(optarg, "esd") == 0)
                {
                    options.opt |= MPG321_USE_ESD;
                }
                
                else
                {
                    fprintf(stderr, "Option %s to --output not implemented or not understood.\n", optarg);
                }
                
                break;    
            
            case 'g':
                options.volume = atoi(optarg);
                break;
            
            case 'V':
                printf("mpg321 version " VERSION "\n"
                       "Copyright (C) 2001 Joe Drew.\n"
                       "This program is free software; you can redistribute it and/or modify\n"
                       "it under the terms of the GNU General Public License as published by\n"
                       "the Free Software Foundation; either version 2 of the License, or\n"
                       "(at your option) any later version.\n" );
                exit(0);
            
            case 'H':
                usage(argv[0]);
                exit(0);
            
            case ':':
                fprintf(stderr, "Missing argument to %s\n", argv[optind]);
                break;
        }
    }
}

int main(int argc, char *argv[])
{
    int fd = 0;
    int already_playing = 0;
    char *currentfile, old_dir[PATH_MAX];
    char * time_formatted;
    playlist *pl = NULL;

    buffer playbuf;
    
    struct mad_decoder decoder;

    /* If there were no arguments just print the usage */
    if (argc == 1)
    {
        usage(argv[0]);
        exit(0);
    }

    old_dir[0] = '\0';

    pl = new_playlist();

    if (!pl)
    {
        fprintf(stderr, "malloc failed at startup!\n");
        exit(1);
    }

    /* Get the command line options */
    options.volume = 100;

    /* also adds all playlist files to playlist pl */
    parse_options(argc, argv, pl);

    if (!(options.opt & MPG321_QUIET_PLAY)) mpg123_boilerplate();

    if (playlist_file)
        load_playlist(pl, playlist_file);
    
    add_cmdline_files(pl, argv);
 
    if (shuffle_play)
        shuffle_files(pl);
        
    ao_initialize();
    
    handle_signals(-1); /* initialize signal handler */
 
    signal(SIGINT, handle_signals);

    /* Play the mpeg files or zip it! */
    while((currentfile = get_next_file(pl)))
    {
        playbuf.buf = NULL;
        playbuf.fd = -1;
        playbuf.length = 0;
        playbuf.done = 0;
        playbuf.num_frames = 0;
        playbuf.filename = currentfile;
        mad_timer_reset(&playbuf.duration);
        
        mad_timer_reset(&current_time);

        /* Create the MPEG stream */
        /* Check if source is on the network */
        if((fd = raw_open(currentfile)) != 0 || (fd = http_open(currentfile)) != 0
            || (fd = ftp_open(currentfile)) != 0)
        {
            playbuf.fd = fd;
            playbuf.buf = malloc(BUF_SIZE);
            mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, /*filter*/0,
                            output, /*error*/0, /* message */ 0);
        }

        /* Check if we are to use stdin for input */
        else if(strcmp(currentfile, "-") == 0)
        {
            playbuf.fd = fileno(stdin);
            playbuf.buf = malloc(BUF_SIZE);
            mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, /*filter*/0,
                            output, /*error*/0, /* message */ 0);
        }
            
        /* currentfile is a local file (presumably.) mmap() it */
        else
        {
            struct stat stat;
            
            if((fd = open(currentfile, O_RDONLY)) == -1)
            {
                fprintf(stderr, "file %s can't be opened\n", currentfile);
                continue;
            }
            
            if(fstat(fd, &stat) == -1)
            {
                fprintf(stderr, "file %s can't be fstat()'ed\n", currentfile);
                continue;
            }
            
            if (!S_ISREG(stat.st_mode))
            {
                fprintf(stderr, "file %s is not a regular file!\n", currentfile);
                continue;
            }
            
            playbuf.length = stat.st_size;
            
            if((playbuf.buf = mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fd, 0))
                                == MAP_FAILED)
            {
                fprintf(stderr, "file %s can't be mmap'ed\n", currentfile);
                continue;
            }

            if (options.opt & MPG321_VERBOSE_PLAY && calc_length(currentfile, &playbuf.duration, &playbuf.num_frames) == -1)
            {
                fprintf(stderr, "Couldn't calculate duration of %s!\n",currentfile);
                continue;
            }
            
            mad_decoder_init(&decoder, &playbuf, read_from_mmap, read_header, /*filter*/0,
                            output, /*error*/0, /* message */ 0);
        }

        if(!(options.opt & MPG321_QUIET_PLAY))/*zip it!!!*/
        {
            /* Because dirname might modify the argument */
            char * dirc = strdup(currentfile);
            char * basec = strdup(currentfile);

            char * basen = basename(basec);
            char * dirn = dirname(dirc);
            

            if(strncmp(old_dir, dirn,PATH_MAX) != 0)
            {
                /* Print information about the file */
                fprintf(stderr, "\n");
                fprintf(stderr,"Directory: %s/\n", dirn);
                
                strncpy(old_dir, dirn, PATH_MAX);
                old_dir[PATH_MAX-1] = '\0';
            }

            /* Print a newline only between new songs */
            if (already_playing)
            {
                fprintf(stderr, "\n");
            }
            else
            {
                already_playing = 1;
            }
            
            fprintf(stderr,"Playing MPEG stream from %s ...\n", basen);
    
            free(dirc);
            free(basec);
        }    

        mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

#if 0
        /* The status line doesn't always equal exactly zero. This should fix that. */
        if (verbose)
        {
            info.current_time = info.total_time;
            print_verbose_status(info);
            printf("\n");
        }
#endif
        
        if (!(options.opt & MPG321_QUIET_PLAY))
        {
            char time_formatted[11];
            mad_timer_string(current_time, time_formatted, "%.2u:%.2u", MAD_UNITS_MINUTES,
                       MAD_UNITS_SECONDS, 0);
            fprintf(stderr, "\n[%s] Decoding of %s finished.\n",time_formatted, basename(currentfile));
        }
        
        mad_decoder_finish(&decoder);

        if (quit_now)
            break;

        if (playbuf.fd == -1)
        {
            munmap(playbuf.buf, playbuf.length);
        }
        else
        {
            free(playbuf.buf);
            if (fd != fileno(stdin)) 
                close(fd);
        }
    }

    if(playdevice)
        ao_close(playdevice);

	ao_shutdown();

#if defined(RAW_SUPPORT) || defined(HTTP_SUPPORT) || defined(FTP_SUPPORT) 
    if(fd) close(fd);
#endif

    return(0);
}
